"""
Builder for jql queries.

Make it easy to programmatically build jira queries.
"""
from __future__ import annotations

from enum import Enum, auto
from typing import Dict, Any, Iterable, Tuple


COMPARE = {"=", "!=", ">", ">=", "<", "<=", "~", "!~"}
COMPARE_LIST = {"IN", "NOT IN"}
CONNECTORS = {"AND", "OR"}
FUNCTIONS = {"linkedissuesof", "membersOf", "issuefieldmatch"}


class _OpType(Enum):
    COMPARE = auto()
    COMPARE_LIST = auto()
    CONNECTOR = auto()
    FUNCTION = auto()
    UNKNOWN = auto()

    @classmethod
    def get_type(cls, symbol: str) -> _OpType:
        symbol = symbol.upper()
        if symbol in COMPARE:
            return _OpType.COMPARE
        if symbol in COMPARE_LIST:
            return _OpType.COMPARE_LIST
        if symbol in CONNECTORS:
            return _OpType.CONNECTOR
        if symbol in FUNCTIONS:
            return _OpType.FUNCTION
        return _OpType.UNKNOWN


def _single_entry(entry: dict) -> Tuple[Any, Any]:
    assert len(entry) == 1
    return [(k, v) for k, v in entry.items()][0]


def _parse_value(value: Any) -> str:
    if isinstance(value, list):
        return f"({','.join(value)})"

    if isinstance(value, dict):
        key, value = _single_entry(value)
        return _parse_function(key, value)

    return str(value)


def _parse_function(key: str, value: Dict) -> str:
    def arg_joiner() -> str:
        return '", "'

    if key.lower() == "linkedissuesof":
        args = [_parse(value["subquery"])]
        if "linktype" in value:
            args.append(value["linktype"])

        return f'linkedIssuesOf("{arg_joiner().join(args)}")'

    if key.lower() == "issuefieldmatch":
        args = [_parse(value["subquery"]), value["field"], value["value"]]
        return f'issueFieldMatch("{arg_joiner().join(args)}")'
    return ""


def _parse_comparison_list(query: Dict) -> str:
    assert len(query) == 1
    operator, values = _single_entry(query)

    return f"{operator} {_parse_value(values)}"


def _parse_comparison(query: Dict) -> str:
    assert len(query) == 1
    operator, value = _single_entry(query)

    return f"{operator} {value}"


def _parse_operator(query: Dict) -> str:
    operator, values = _single_entry(query)
    key_type = _OpType.get_type(operator)
    if key_type == _OpType.COMPARE:
        return _parse_comparison(query)
    if key_type == _OpType.COMPARE_LIST:
        return _parse_comparison_list(query)
    return ""


def _parse_field_search(field: str, search: Any) -> str:
    if isinstance(search, dict):
        return f"'{field}' {_parse_operator(search)}"
    return f"'{field}' {_parse_value(search)}"


def _parse_connector(connector: str, values: Iterable) -> str:
    return connector.upper().join([f" ({_parse(item)}) " for item in values])


def _parse_item(key: str, value: Any) -> str:
    key_type = _OpType.get_type(key)
    if key_type == _OpType.COMPARE:
        return _parse_comparison({key: value})
    if key_type == _OpType.CONNECTOR:
        return _parse_connector(key, value)
    return _parse_field_search(key, value)


def _parse(query: Dict) -> str:
    return " ".join([_parse_item(k, v) for k, v in query.items()])


def jql_from_dict(query_dict: Dict) -> str:
    """
    Convert the dictionary to a jql query.

    :param query_dict: Dictionary containing query.
    :return: jql query.
    """
    return _parse(query_dict)



"""Wrapper for jira object."""
from __future__ import annotations

from typing import Dict, Optional, Any, Iterable

from jira import JIRA, Issue

from herakles.auth import JiraAuth
from herakles.jql.jql_builder import jql_from_dict
from herakles.util.file_utils import read_yaml_file

DEFAULT_NETWORK_TIMEOUT = 120
DEFAULT_LABEL = "jira_custom_fields"


class JiraWrapper(object):
    """Make calls to Jira."""

    def __init__(self, jira: JIRA, custom_field_map: Optional[Dict] = None):
        """
        Create a wrapper for Jira API.

        :param jira: Jira client to wrap.
        :param custom_field_map: Dictionary mapping custom fields.
        """
        self.jira = jira
        self._custom_field_map = custom_field_map

    @classmethod
    def connect(
        cls,
        jira_server: str,
        auth: JiraAuth,
        network_timeout: int = DEFAULT_NETWORK_TIMEOUT,
        custom_field_map: Optional[Dict] = None,
    ) -> JiraWrapper:
        """
        Connect to the specified Jira instance.

        :param jira_server: Hostname of jira instance.
        :param auth: Authentication information.
        :param network_timeout: Seconds until network timeout.
        :param custom_field_map: Dictionary with mapping of custom fields.
        :return: Wrapper to connect with Jira.
        """
        connect_args = {
            "options": {"server": jira_server},
            "validate": True,
            "timeout": network_timeout,
        }
        connect_args.update(auth.auth_dict())
        jira = JIRA(**connect_args)
        return cls(jira, custom_field_map)

    def add_custom_fields_from_file(self, file_path: str, label: str = DEFAULT_LABEL) -> None:
        """
        Add a mapping of custom fields from a yaml file.

        :param file_path: Yaml file containing custom fields.
        :param label: Key the custom fields are under.
        """
        self._custom_field_map = read_yaml_file(file_path)[label]

    def get_issue(self, jira_issue: str) -> IssueWrapper:
        """
        Retrieve the given jira issue.

        :param jira_issue: Jira issue to query.
        :return: Jira Issue.
        """
        return IssueWrapper(self.jira.issue(jira_issue), self._custom_field_map)

    def search_issues(self, search: Dict, **kwargs: Optional[Dict]) -> Iterable[IssueWrapper]:
        """
        Search for jira issues.

        :param search: Dictionary specifying search parameters.
        :return: Iterable of issues found.
        """
        jql = jql_from_dict(search)
        results = self.jira.search_issues(jql, **kwargs)
        for issue in results:
            yield IssueWrapper(issue, self._custom_field_map)

    def sprints_by_name(self, name: str) -> Dict:
        """
        Get sprint with the given name.

        :param name: Name of sprint to get.
        :return: Sprint with given name.
        """
        return self.jira.sprints_by_name(name)


class IssueWrapper(object):
    """Jira issue."""

    def __init__(self, jira_issue: Issue, custom_field_map: Optional[Dict] = None):
        """
        Create an IssueWrapper for the given issue.

        :param jira_issue: Issue from jira client.
        :param custom_field_map: Dictionary with mappings of custom fields.
        """
        self._issue = jira_issue
        if custom_field_map is None:
            custom_field_map = {}
        self._custom_field_map = custom_field_map

    def __getattr__(self, item: str) -> Any:
        """
        Lookup an attribute on the given issue.

        :param item: attribute to lookup.
        :return: Value of attribute.
        """
        if item in self._custom_field_map:
            return getattr(self, self._custom_field_map[item])

        if hasattr(self._issue, item):
            return getattr(self._issue, item)

        return getattr(self._issue.fields, item)

    @property
    def key(self) -> str:
        """Jira key of issue."""
        return self._issue.key



"""Authentication information."""
import abc
from enum import Enum, auto
from typing import Dict, Tuple, Optional

from herakles.util.file_utils import read_yaml_file

DEFAULT_AUTH_KEY = "jira"


class AuthType(Enum):
    """Type of authentication to use."""

    BASIC = auto()
    OAUTH = auto()


class JiraAuth(abc.ABC):
    """Container for Jira authentication."""

    @abc.abstractmethod
    def auth_type(self) -> AuthType:
        """Get the type of authentication to use."""
        raise NotImplementedError

    @abc.abstractmethod
    def auth_dict(self) -> Dict:
        """Get a dictionary for authentication."""
        raise NotImplementedError


class JiraAuthBasic(JiraAuth):
    """Basic Authentication for Jira."""

    def __init__(self, username: str, password: Optional[str] = None):
        """
        Create a basic auth configuration for jira.

        :param username: Username to use.
        :param password: Password to use.
        """
        super().__init__()
        self.username = username
        self.password = password

    def auth_type(self) -> AuthType:
        """Get the type of authentication to use."""
        return AuthType.BASIC

    def _basic_auth(self) -> Tuple[str, Optional[str]]:
        """Return a user, password tuple that can be used to configure JIRA basic auth."""
        return self.username, self.password

    def auth_dict(self) -> Dict:
        """Get a dictionary for authentication."""
        return {"basic_auth": self._basic_auth()}


class JiraAuthOAuth(JiraAuth):
    """OAuth configuration for Jira."""

    def __init__(
        self, access_token: str, access_token_secret: str, consumer_key: str, key_cert: str
    ):
        """
        Create an OAuth configuration for Jira.

        :param access_token: Access token to use.
        :param access_token_secret: Access token secret to use.
        :param consumer_key: Consumer key to use.
        :param key_cert: Key certificate to use.
        """
        super().__init__()
        self.access_token = access_token
        self.access_token_secret = access_token_secret
        self.consumer_key = consumer_key
        self.key_cert = key_cert

    @classmethod
    def from_dict(cls, auth_config: Dict) -> JiraAuth:
        """
        Create a JiraAuth object for OAuth from a dictionary.

        The dictionary should contain the following: 'access_token', 'access_token_secret',
        'consumer_key', and 'key_cert'.

        :param auth_config: Dictionary containing authentication information.
        :return: JiraAuth object for OAuth.
        """
        return cls(
            auth_config["access_token"],
            auth_config["access_token_secret"],
            auth_config["consumer_key"],
            auth_config["key_certificate"],
        )

    @classmethod
    def from_yaml_file(cls, yaml_file: str, config_key: str = DEFAULT_AUTH_KEY) -> JiraAuth:
        """
        Create a JiraAuth object for OAuth from a yaml file.

        The file should contain the following under the specified key: 'access_token',
        'access_token_secret', 'consumer_key', and 'key_certificate'.

        :param yaml_file: Path to yaml file.
        :param config_key: Top level configuration key.
        :return: JiraAuth object.
        """
        config = read_yaml_file(yaml_file)
        return cls.from_dict(config[config_key])

    def auth_type(self) -> AuthType:
        """Get the type of authentication to use."""
        return AuthType.OAUTH

    def _get_oauth(self) -> Dict:
        """Return a dictionary that can be used to configure JIRA oauth."""
        return {
            "access_token": self.access_token,
            "access_token_secret": self.access_token_secret,
            "consumer_key": self.consumer_key,
            "key_cert": self.key_cert,
        }

    def auth_dict(self) -> Dict:
        """Get a dictionary for authentication."""
        return {"oauth": self._get_oauth()}



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['herakles', 'herakles.jql', 'herakles.util']

package_data = \
{'': ['*']}

install_requires = \
['PyYaml>=5.0,<6.0', 'jira>=2.0,<3.0']

setup_kwargs = {
    'name': 'herakles',
    'version': '0.1.0',
    'description': '',
    'long_description': '# herakles jira tool\n\nTools for working with the [Jira API](https://jira.readthedocs.io/en/master/index.html).\n\n\n## Usage\n\n### Initialize\n\n```python\nfrom herakles import JiraAuthBasic, JiraWrapper\nauth = JiraAuthBasic("username", "password")\njira = JiraWrapper.connect("https://server.jira.com", auth)\n\n# optional add map of custom fields.\njira.add_custom_fields_from_file("custom_fields.yml")\n```\n\n### Use jql builder to programmatically build jql queries\n\nSee how many tickets a user resolved in the past number days:\n```python\nquery = {\n            "and": [\n                {"assignee": {"=": team_member}},\n                {"resolved": {">": f"-{days}d"}},\n                {"project": {"not in": ["\'Unrelated Project\'"]}},\n            ]\n        }\n```\n\n## Contributors Guide\n\n### Testing\n\nTesting is done via pytest.\n\n```\n$ pip install -r requirements.txt\n$ pytest\n```\n\nTo get code coverage information, you can run pytest directly.\n\n```\n$ pip install -r requirements.txt\n$ pytest --cov=src --cov-report=html\n```\n\nThis will generate an html coverage report in `htmlcov/` directory.\n',
    'author': 'David Bradford',
    'author_email': 'david.bradford@mongodb.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dbradf/herakles-jira-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Various utilities for working with files."""
from typing import Dict

import yaml


def read_yaml_file(file_path: str) -> Dict:
    """
    Read a yaml file and return the contents as a dictionary.

    :param file_path: Path to yaml files.
    :return: Dictionary with file contents.
    """
    with open(file_path) as file_contents:
        return yaml.safe_load(file_contents)
